# amaranth: UnusedElaboratable=no

import pytest
from smolarith.mul import MulticycleMul, PipelinedMul, Sign
from collections import deque
from itertools import product
import random


def make_testbench(mod, values):
    m = mod

    if isinstance(mod, PipelinedMul):
        async def tb(ctx):
            # Pipeline previous inputs... inputs to prev.append() are what
            # just went into the multiplier at the current active edge.
            # Outputs from prev.popleft() are what went into the multiplier
            # "m.width" active edges ago. This leads to a latency of
            # "m.width" clock cycles/ticks since the multiplier saw the inputs.
            #
            # We only need m.width - 1 storage, because a latency of "1" means
            # the multiplier outputs will be immediately available after the
            # current active edge that just happened (no need to store!).
            # This testbench doesn't work with m.width == 1 :).
            prev = deque([(0, 0)]*(m.width - 1))

            # Not yielding to the simulator when values don't change
            # can result in significant speedups.
            a_prev = None
            b_prev = None
            s_prev = None

            await ctx.tick()

            ctx.set(m.inp.valid, 1)
            ctx.set(m.outp.ready, 1)

            for (a, b, s) in values:
                if s == Sign.UNSIGNED:
                    if a != a_prev:
                        ctx.set(m.inp.payload.a, a)
                    if b != b_prev:
                        ctx.set(m.inp.payload.b, b)
                else:
                    if a != a_prev:
                        ctx.set(m.inp.payload.a.as_signed(), a)
                    if b != b_prev:
                        ctx.set(m.inp.payload.b.as_signed(), b)

                if s != s_prev:
                    ctx.set(m.inp.payload.sign, s)
                await ctx.tick()
                (a_prev, b_prev, s_prev) = (a, b, s)

                (a_c, b_c) = prev.popleft()
                prev.append((a, b))

                if ctx.get(m.outp.payload.sign) == Sign.UNSIGNED.value:
                    assert a_c*b_c == ctx.get(m.outp.payload.o)
                else:
                    assert a_c*b_c == ctx.get(m.outp.payload.o.as_signed())

                # print((a, b), (a_c, b_c), a_c*b_c, (yield m.o))
                # for i in range(8):
                #     print(f"{yield m.pin[i].a:08b}, {yield m.pin[i].b:08b}")
                # for i in range(8):
                #      print(f"{yield m.pout[i]:016b}")

            # Drain pipeline.
            for _ in range(m.width):
                await ctx.tick()
                (a_c, b_c) = prev.popleft()
                prev.append((a, b))

                if ctx.get(m.outp.payload.sign) == Sign.UNSIGNED.value:
                    assert a_c*b_c == ctx.get(m.outp.payload.o)
                else:
                    assert a_c*b_c == ctx.get(m.outp.payload.o.as_signed())
    elif isinstance(mod, MulticycleMul):
        async def tb(ctx):
            # Not yielding to the simulator when values don't change
            # can result in significant speedups.
            a_prev = None
            b_prev = None
            s_prev = None

            await ctx.tick()

            ctx.set(m.inp.valid, 1)
            ctx.set(m.outp.ready, 1)

            for (a, b, s) in values:
                if s == Sign.UNSIGNED:
                    if a != a_prev:
                        ctx.set(m.inp.payload.a, a)
                    if b != b_prev:
                        ctx.set(m.inp.payload.b, b)
                else:
                    if a != a_prev:
                        ctx.set(m.inp.payload.a.as_signed(), a)
                    if b != b_prev:
                        ctx.set(m.inp.payload.b.as_signed(), b)

                if s != s_prev:
                    ctx.set(m.inp.payload.sign, s)
                ctx.set(m.inp.valid, 1)
                await ctx.tick()
                (a_prev, b_prev, s_prev) = (a, b, s)

                ctx.set(m.inp.valid, 0)
                await ctx.tick().until(m.outp.valid == 1)

                if ctx.get(m.outp.payload.sign) == Sign.UNSIGNED.value:
                    assert a*b == ctx.get(m.outp.payload.o)
                else:
                    assert a*b == ctx.get(m.outp.payload.o.as_signed())
    else:
        assert False

    return tb


@pytest.fixture(params=[Sign.UNSIGNED, Sign.SIGNED, Sign.SIGNED_UNSIGNED],
                ids=["u", "i", "iu"])
def all_values(request, mod):
    m = mod
    mode = request.param

    if mode == Sign.UNSIGNED:
        a_range = range(0, 2**m.width)
        b_range = range(0, 2**m.width)
    elif mode == Sign.SIGNED:
        a_range = range(-2**(m.width-1), 2**(m.width-1))
        b_range = range(-2**(m.width-1), 2**(m.width-1))
    else:  # Sign.SIGNED_UNSIGNED
        a_range = range(-2**(m.width-1), 2**(m.width-1))
        b_range = range(0, 2**m.width)

    return product(a_range, b_range, (mode,))


@pytest.fixture
def random_vals(mod):
    w = mod.width -1

    def shift_to_unsigned(v):
        if v == -2**w:
            return 0  # Most negative value maps to 0.
        elif v < 0:
            return v * -1
        else:  # v >= 0
            return v | (1 << 31)

    def vals():
        for i in range(256):
            a = random.randint(-2**w, (2**w)-1)
            b = random.randint(-2**w, (2**w)-1)
            s = random.choice([Sign.UNSIGNED, Sign.SIGNED,
                               Sign.SIGNED_UNSIGNED])

            if s == Sign.UNSIGNED:
                a = shift_to_unsigned(a)
                b = shift_to_unsigned(b)
            elif s == Sign.SIGNED_UNSIGNED and b < 0:
                b = shift_to_unsigned(b)

            yield (a, b, s)

    return vals()


@pytest.fixture
def pipeline_tb(mod):
    async def testbench(ctx):
        m = mod

        await ctx.tick()

        ctx.set(m.inp.payload.sign, Sign.UNSIGNED)
        ctx.set(m.inp.payload.a, 1)
        ctx.set(m.inp.payload.b, 1)
        ctx.set(m.inp.valid, 1)
        ctx.set(m.outp.ready, 1)
        await ctx.tick()

        ctx.set(m.inp.payload.a, 2)
        ctx.set(m.inp.payload.b, 2)
        await ctx.tick()

        # Pipeline should continue working...
        ctx.set(m.inp.valid, 0)
        await ctx.tick()

        ctx.set(m.inp.valid, 1)
        ctx.set(m.inp.payload.a, 3)
        ctx.set(m.inp.payload.b, 3)
        await ctx.tick()

        ctx.set(m.inp.payload.a, 4)
        ctx.set(m.inp.payload.b, 4)
        await ctx.tick()

        ctx.set(m.inp.valid, 0)
        await ctx.tick().repeat(3)

        ctx.set(m.outp.ready, 0)
        assert ctx.get(m.outp.valid) == 1
        assert ctx.get(m.outp.payload.sign) == Sign.UNSIGNED
        assert ctx.get(m.outp.payload.o) == 1
        # Pipeline should stall...
        assert ctx.get(m.inp.ready) == 0
        await ctx.tick()

        # Until we indicate we're ready to accept data.
        ctx.set(m.outp.ready, 1)
        assert ctx.get(m.inp.ready) == 1
        await ctx.tick()

        assert ctx.get(m.outp.valid) == 1
        assert ctx.get(m.outp.payload.sign) == Sign.UNSIGNED
        assert ctx.get(m.outp.payload.o) == 4
        assert ctx.get(m.inp.ready) == 1
        await ctx.tick()

        assert ctx.get(m.outp.valid) == 0
        assert ctx.get(m.inp.ready) == 1
        await ctx.tick()

        assert ctx.get(m.outp.valid) == 1
        assert ctx.get(m.outp.payload.sign) == Sign.UNSIGNED
        assert ctx.get(m.outp.payload.o) == 9
        assert ctx.get(m.inp.ready) == 1
        await ctx.tick()

        assert ctx.get(m.outp.valid) == 1
        assert ctx.get(m.outp.payload.sign) == Sign.UNSIGNED
        assert ctx.get(m.outp.payload.o) == 16
        assert ctx.get(m.inp.ready) == 1
        await ctx.tick()

        assert ctx.get(m.outp.valid) == 0
        assert ctx.get(m.inp.ready) == 1
        await ctx.tick()

    return testbench


@pytest.mark.parametrize("mod", [PipelinedMul(8, debug=True),
                                 MulticycleMul(6)])
@pytest.mark.parametrize("clks", [1.0 / 12e6])
def test_all_values(sim, mod, all_values):
    sim.run(testbenches=[make_testbench(mod, all_values)])


@pytest.mark.parametrize("mod", [PipelinedMul(32, debug=True),
                                 MulticycleMul(32),
                                 PipelinedMul(64, debug=True),
                                 MulticycleMul(64)],
                         ids=["p32", "m32", "p64", "m64"])
@pytest.mark.parametrize("clks", [1.0 / 12e6])
def test_random(sim, mod, random_vals):
    random.seed(0)
    sim.run(testbenches=[make_testbench(mod, random_vals)])


@pytest.mark.parametrize("mod,clks", [(PipelinedMul(8, debug=True),
                                       1.0 / 12e6)])
def test_pipeline_stall(sim, pipeline_tb):
    sim.run(testbenches=[pipeline_tb])



"""Soft-core divider components."""

from amaranth import Elaboratable, Module, Signal, signed, C, unsigned
from amaranth.lib.data import StructLayout
from amaranth.lib.wiring import In, Out, Component
from amaranth.lib.enum import Enum, auto
from amaranth.lib import stream


class Sign(Enum):
    """Indicate the type of divide to be performed.

    Attributes
    ----------
    UNSIGNED : int
        Both inputs ``n`` and ``d`` are unsigned.

        The output is unsigned.

    SIGNED : int
        Both inputs ``n`` and ``d`` are signed.

        The quotient and remainder are signed. The remainder takes
        the sign of the dividend.
    """

    UNSIGNED = auto()
    SIGNED = auto()


class Inputs(StructLayout):  # noqa: DOC602,DOC603
    r"""Tagged union representing the signedness of divider inputs.

    * When :attr:`sign` is ``UNSIGNED``, both :attr:`n` and :attr:`d` are
      treated by the divider as :class:`~amaranth:amaranth.hdl.Value`\ s
      with an :data:`~amaranth:amaranth.hdl.unsigned`
      :class:`~amaranth:amaranth.hdl.Shape`.

    * When ``sign`` is ``SIGNED``, both ``n`` and ``d`` are treated as Values
      with a :data:`~amaranth:amaranth.hdl.signed` Shape.

    Parameters
    ----------
    width : int
        Width in bits of both inputs ``n`` and ``d``. For signed divides, this
        includes the sign bit.

    Attributes
    ----------
    sign: Sign
        Controls the interpretation of the bit patterns of ``n`` and ``d``
        during division.
    n: Signal(width)
        The dividend; i.e. the ':math:`n`' in :math:`n / d`.
    d: Signal(width)
        The divisor; i.e. the ':math:`d`' in :math:`n / d`.
    """

    def __init__(self, width):
        super().__init__({
            "sign": Sign,
            "n": unsigned(width),
            "d": unsigned(width),
        })


class Outputs(StructLayout):  # noqa: DOC602,DOC603
    r"""Tagged union representing the signedness of divider outputs.

    * When :attr:`sign` is ``UNSIGNED``, :attr:`q` and :attr:`r` should be
      treated as :class:`~amaranth:amaranth.hdl.Value`\ s with an
      :class:`~amaranth:amaranth.hdl.as_unsigned`
      :class:`~amaranth:amaranth.hdl.Shape`.

    * When ``sign`` is ``SIGNED``, ``q`` and ``r`` should be treated as Values
      with a :class:`~amaranth:amaranth.hdl.as_signed` Shape.

    Parameters
    ----------
    width : int
        Width in bits of the outputs ``q`` and ``r``. For signed dividers, this
        includes the sign bit.

    Attributes
    ----------
    sign: Sign
        Indicates whether the divider that produced this quotient/remainder
        was signed or unsigned.
    q: Signal(width)
        The quotient of :math:`n / d`.
    r: Signal(width)
        The remainder of :math:`n / d`, i.e. :math:`n \bmod d`.
    """

    def __init__(self, width):
        super().__init__({
            "sign": Sign,
            "q": unsigned(width),
            "r": unsigned(width)
        })


def divider_input_signature(width):
    """Create a parametric divider input port.

    This function returns a :class:`~amaranth:amaranth.lib.stream.Signature`
    that's usable as a transfer initiator to a divider. A divider starts
    on the current cycle when both ``valid`` and ``rdy`` are asserted.

    Parameters
    ----------
    width : int
        Width in bits of the inputs :attr:`n` and :attr:`d`. For signed
        divides, this includes the sign bit.

    Returns
    -------
    :class:`amaranth:amaranth.lib.stream.Signature`
        :py:`Signature(Inputs)`
    """
    return stream.Signature(Inputs(width))


def divider_output_signature(width):
    """Create a parametric divider output port.

    This function returns a :class:`~amaranth:amaranth.lib.stream.Signature`
    that's usable as a transfer initiator **from** a divider.

    .. note:: For a core responding **to** a divider, which is the typical
              use case, you will want to use this Signature with the
              :data:`~amaranth:amaranth.lib.wiring.In` flow, like so:

              .. doctest::

                  >>> from smolarith.div import divider_output_signature
                  >>> from amaranth.lib.wiring import Signature, In
                  >>> my_receiver_sig = Signature({
                  ...     "inp": In(divider_output_signature(width=8))
                  ... })

    Parameters
    ----------
    width : int
        Width in bits of the outputs ``q`` and ``r``. For signed divides, this
        includes the sign bit.

    Returns
    -------
    :class:`amaranth:amaranth.lib.stream.Signature`
        :py:`Signature(Outputs)`
    """
    return stream.Signature(Outputs(width))


class _Quadrant(StructLayout):
    """Store sign information about divider inputs."""

    def __init__(self):  # noqa: DOC1
        super().__init__({
            "n": unsigned(1),
            "d": unsigned(1),
        })


class _Negate(Component):
    def __init__(self, width=8):  # noqa: DOC1
        self.width = width
        super().__init__({
            "inp": In(self.width),
            "outp": Out(self.width)
        })

    
    def elaborate(self, plat):
        m = Module()

        m.d.comb += self.outp.eq((-self.inp.as_signed())[:-1])

        return m


class Impl(Enum):
    """Indicate the division algorithm to use inside a dividing component.

    Attributes
    ----------
    RESTORING : int
        Use :ref:`restoring division <derive-res>` for the internal divider.
        *This is a good default*.

    NON_RESTORING : int
        Use :ref:`non-restoring division <derive-nr>` for the internal divider.
    """

    RESTORING = auto()
    NON_RESTORING = auto()


class MulticycleDiv(Component):  # noqa: DOC602,DOC603
    # FIXME: I can't be .. _latency label to work, even with :ref:`latency`...
    # always "undefined label"... huh?!
    r"""Restoring/Non-restoring divider soft-core.

    This divider core is a gateware implementation of 
    :ref:`restoring division <derive-res>` or
    :ref:`non-restoring division <derive-nr>`. It works for both signed and
    unsigned divides, and should be preferred to :class:`LongDivider` in almost
    all circumstances due to better resource usage.

    * A divide starts on the current cycle when both ``inp.valid`` and
      ``inp.ready`` are asserted.
    * A divide result is available when ``outp.valid`` is asserted. The
      result is read/overwritten once a downstream core asserts ``outp.ready``.

    * Latency: Divide Results for a restoring divider will be available
      ``width + 3`` clock cycles after assertion of both ``inp.valid`` and
      ``inp.ready``. For a non-restoring divider, results are available after
      ``width + 4`` clock cycles.

    * Throughput: One divide maximum is finished every ``width + 3``
      clock cycles for a restoring divider, and every ``width + 4`` for a
      non-restoring divider.

    Parameters
    ----------
    width : int
        Width in bits of both inputs ``n`` and ``d``. For signed
        multiplies, this includes the sign bit. Outputs ``q`` and ``r`` width
        will be the same width.
    impl : Impl
        Choose whether to use a restoring divider or non-restoring divider
        internally. The default of ``RESTORING`` is fine for most cases.

    Attributes
    ----------
    width : int
        Bit width of the inputs ``n`` and ``d``, and outputs ``q`` and ``r``.
    inp : In(divider_input_signature(width))
        Input interface to the divider.
    outp : Out(divider_output_signature(width))
        Output interface of the divider.

    Notes
    -----
    * This divider is implemented using either restoring or non-restoring
      division. It is basically a gateware translation of the
      :ref:`Python <resdiv-py>` :ref:`implementations <nrdiv-py>` shown in
      :ref:`impl`.

    * Unlike the multipliers, where I could eyeball approximate storage element
      usage, divider size was benchmarked using
      `yosys <https://yosyshq.net/yosys/>`_\ 's ``synth_ice40`` pass. For an
      :math:`n`-bit divide:
    
      * ``impl=Impl.RESTORING`` requires approximately :math:`O(7*n)`
        storage elements and :math:`O(12.75*n)` LUT4s.
      * ``impl=Impl.NON_RESTORING`` requires approximately :math:`O(8*n)`
        storage elements and :math:`O(14.5*n)` LUT4s.

    * Internally, the divider converts its operands from signed to unsigned
      if necessary, performs the division, and the converts back from unsigned
      to signed if necessary. I ran into some :ref:`issues <signedness>` with
      making a signed-aware non-restoring divider such that eating the
      conversion cost latency is probably justifiable for implementation
      simplicity of both a restoring and non-restoring divider.

    .. _latency:

      The combination of signed-to-unsigned conversion, using a cycle to
      latching inputs to the internal divider, and unsigned-to-signed
      conversion adds three cycles of latency beyond the expected ``width``
      number of cycles to perform a division.

      Additionally, when using a non-restoring divider, the quotient and
      remainder require a possible :ref:`final restore step <nrdiv-restore>`.
      Remainder restore is implemented as shown in the
      :ref:`Python code <nrdiv-py>`. Quotient restore is implemented by
      subtracting the ones complement of the raw quotient from the raw quotient
      itself.

    * The quotient and remainder share a backing store; new bits are shifted
      into the lower-half quotient portion as the remainder is shifted into the
      upper half. This works because each iteration of the
      restoring/non-restoring loops check the sign of the remainder, and the
      lower quotient bits won't affect the sign bit.

    * This divider is compliant with RISC-V semantics for divide-by-zero and
      overflow when ``width=32`` or ``width=64``\ [rv]_. Specifically:

      * Signed/unsigned divide-by-zero returns "all bits set" for the
        quotient and the dividend as the remainder.
      * Signed overflow (:math:`-2^{\{31,63\}}/-1`) returns
        :math:`-2^{\{31,63\}}` for the quotient and :math:`0` as the
        remainder.

    Future Directions
    -----------------

    * It may be worthwhile to make a pipelined divider?

    * Current latency is worse than the :class:`LongDivider`, which is just
      about the only advantage to ``LongDivider``. We can provide an option to
      reduce latency/resources required due to the signed/unsigned and restore
      stages at the cost of some timing closure.
    """

    def __init__(self, width=8, impl=Impl.RESTORING):
        self.width = width
        if impl == Impl.RESTORING:
            self.div_impl = _RestoringDiv(width)
        elif impl == Impl.NON_RESTORING:
            self.div_impl = _NonRestoringDiv(width)
        self.negate_nq = _Negate(width)
        self.negate_dr = _Negate(width)

        super().__init__({
            "inp": In(divider_input_signature(self.width)),
            "outp": Out(divider_output_signature(self.width))
        })

    def elaborate(self, plat):  # noqa: D102
        m = Module()

        m.submodules.div_impl = self.div_impl
        # Time-division multiplex negation units... in a multicycle div, only
        # n/d or q/r need to be negated at once.
        m.submodules.negate_nq = self.negate_nq
        m.submodules.negate_dr = self.negate_dr

        quad = Signal(_Quadrant())
        div_inputs = Signal(Inputs(self.width))

        m.d.comb += self.div_impl.inp.payload.eq(div_inputs)

        with m.FSM():
            with m.State("IDLE"):
                m.d.comb += [
                    self.inp.ready.eq(1),
                    self.negate_nq.inp.eq(self.inp.payload.n),
                    self.negate_dr.inp.eq(self.inp.payload.d)
                ]

                m.d.comb += self.inp.ready.eq(1)
                with m.If(self.inp.valid):
                    m.next = "DIVIDE_IN"
                    
                    # Re: (self.inp.payload.d != 0)
                    # For RISCV compliance when dividing by zero, we need to
                    # suppress signed-unsigned conversion. Treating the input
                    # bit pattern as-is will result in the correct behavior of
                    # "all bits set in divisor" and remainder unchanged,
                    # regardless of sign.
                    m.d.sync += [
                        quad.n.eq(self.inp.payload.n[-1] &
                                  (self.inp.payload.d != 0)),
                        quad.d.eq(self.inp.payload.d[-1]),
                        div_inputs.eq(self.inp.payload)
                    ]

                    with m.If(self.inp.payload.sign == Sign.SIGNED):
                        with m.If(self.inp.payload.n[-1] &
                                  (self.inp.payload.d != 0)):
                            m.d.sync += div_inputs.n.eq(self.negate_nq.outp)
                        with m.If(self.inp.payload.d[-1]):
                            m.d.sync += div_inputs.d.eq(self.negate_dr.outp)

            with m.State("DIVIDE_IN"):
                m.d.comb += self.div_impl.inp.valid.eq(1)

                with m.If(self.div_impl.inp.ready):
                    m.next = "DIVIDE_LOOP"

            with m.State("DIVIDE_LOOP"):
                m.d.comb += [
                    self.div_impl.outp.ready.eq(1),
                    self.negate_nq.inp.eq(self.div_impl.outp.payload.q),
                    self.negate_dr.inp.eq(self.div_impl.outp.payload.r)
                ]

                with m.If(self.div_impl.outp.valid):
                    m.next = "DIVIDE_OUT"

                    m.d.sync += [
                        self.outp.payload.eq(self.div_impl.outp.payload),
                    ]

                    with m.If((div_inputs.sign == Sign.SIGNED) &
                              quad.n & ~quad.d):
                        m.d.sync += [
                            self.outp.payload.q.eq(self.negate_nq.outp),
                            self.outp.payload.r.eq(self.negate_dr.outp)
                        ]

                    with m.If((div_inputs.sign == Sign.SIGNED) &
                              ~quad.n & quad.d):
                        m.d.sync += self.outp.payload.q.eq(self.negate_nq.outp)
                        
                    with m.If((div_inputs.sign == Sign.SIGNED) &
                              quad.n & quad.d):
                        m.d.sync += self.outp.payload.r.eq(self.negate_dr.outp)
                        
            with m.State("DIVIDE_OUT"):
                m.d.comb += self.outp.valid.eq(1)

                with m.If(self.outp.ready):
                    m.next = "IDLE"

        return m


class _RestoringDiv(Component):  # noqa: DOC602,DOC603
    """Unsigned-only restoring divider."""

    def __init__(self, width=8):  # noqa: DOC1
        self.width = width
        # sign Signals are unused- Sign.UNSIGNED is implicit.
        super().__init__({
            "inp": In(divider_input_signature(self.width)),
            "outp": Out(divider_output_signature(self.width))
        })

    def elaborate(self, platform):
        m = Module()

        # Need extra bit because we need to subtract up to 2^width, which
        # should remain negative.
        intermediate = Signal(2*self.width + 1)
        iters_left = Signal(range(self.inp.payload.n.shape().width + 1))
        in_progress = Signal()
        s_copy = Signal(Sign)
        d_copy = Signal(self.inp.payload.d.shape())

        m.d.comb += self.inp.ready.eq((self.outp.ready & self.outp.valid) |
                                      ~in_progress)
        m.d.comb += in_progress.eq(iters_left != 0)

        with m.If(self.outp.valid & self.outp.ready):
            m.d.sync += self.outp.valid.eq(0)

        with m.If(self.inp.ready & self.inp.valid):
            m.d.sync += [
                s_copy.eq(self.inp.payload.sign),
                d_copy.eq(self.inp.payload.d),
                iters_left.eq(self.inp.payload.n.shape().width),
                intermediate.eq(self.inp.payload.n)
            ]

        with m.If(in_progress):
            # State Control
            m.d.sync += iters_left.eq(iters_left - 1)

            with m.If((iters_left - 1) == 0):
                m.d.sync += self.outp.valid.eq(1)

            # Loop iter
            # S = (S << 1) - D
            with m.If(((intermediate << 1) - (d_copy << self.width)) < 0):
                # S += D
                m.d.sync += intermediate.eq((intermediate << 1))  # noqa: E501
            with m.Else():
                # q <<=1 and then q |= 1
                # (Docs do q |= 1 and then q <<= 1, requiring a >> at the end.)
                m.d.sync += intermediate.eq(((intermediate << 1) -
                                             (d_copy << self.width)) | 1)  # noqa: E501

        m.d.comb += [
            self.outp.payload.q.eq(intermediate[:self.width]),
            self.outp.payload.r.eq(intermediate[self.width:]),
            self.outp.payload.sign.eq(s_copy)
        ]

        return m


class _NonRestoringDiv(Component):
    """Unsigned-only non-restoring divider."""

    def __init__(self, width=8):  # noqa: DOC1
        self.width = width
        # sign Signals are unused- Sign.UNSIGNED is implicit.
        super().__init__({
            "inp": In(divider_input_signature(self.width)),
            "outp": Out(divider_output_signature(self.width))
        })

    def elaborate(self, platform):
        m = Module()

        # Need extra bit because we need to subtract up to 2^width, which
        # should remain negative.
        intermediate = Signal(2*self.width + 1)
        iters_left = Signal(range(self.inp.payload.n.shape().width + 1))
        in_progress = Signal()
        restore_step = Signal()
        s_copy = Signal(Sign)
        d_copy = Signal(self.inp.payload.d.shape())
        r = Signal(self.width)

        m.d.comb += self.inp.ready.eq((self.outp.ready & self.outp.valid) |
                                      ~in_progress)
        m.d.comb += in_progress.eq((iters_left != 0) | restore_step)

        with m.If(self.outp.valid & self.outp.ready):
            m.d.sync += self.outp.valid.eq(0)

        with m.If(self.inp.ready & self.inp.valid):
            m.d.sync += [
                s_copy.eq(self.inp.payload.sign),
                d_copy.eq(self.inp.payload.d),
                iters_left.eq(self.inp.payload.n.shape().width),
                intermediate.eq(self.inp.payload.n)
            ]

        with m.If(in_progress & ~restore_step):
            # State Control
            m.d.sync += iters_left.eq(iters_left - 1)

            with m.If((iters_left - 1) == 0):
                m.d.sync += restore_step.eq(1)

            # Loop iter
            with m.If(intermediate[-1]):
                # S = (S << 1) + D, encoded in top half.
                # q = -1, encoded as 0, encoded in bottom half.
                m.d.sync += intermediate.eq(((intermediate << 1) +
                                            (d_copy << self.width)))  # noqa: E501
            with m.Else():
                # S = (S << 1) - D, encoded in top half.
                # q = 1, encoded as 1, encoded in bottom half.
                m.d.sync += intermediate.eq(((intermediate << 1) -
                                            (d_copy << self.width)) | 1)  # noqa: E501


        with m.If(restore_step):
            m.d.sync += [
                restore_step.eq(0),
                self.outp.valid.eq(1)
            ]

            # Extract encoded negative powers of two and then subtract as if
            # they were positive powers of two (so no twos complement
            # conversion needed).
            # Better LUT usage if "r" has its own register.
            m.d.sync += [
                intermediate[:self.width].eq(
                    intermediate[:self.width] - ~intermediate[:self.width]),
                r.eq(intermediate[self.width:])
            ]

            with m.If(intermediate[-1]):
                # S += D
                # q -= 1
                m.d.sync += [
                    intermediate[:self.width].eq(
                        intermediate[:self.width] -
                        ~intermediate[:self.width] -
                        1),
                    r.eq(intermediate[self.width:] + d_copy)
                ]

        m.d.comb += [
            self.outp.payload.q.eq(intermediate[:self.width]),
            self.outp.payload.r.eq(r),
            self.outp.payload.sign.eq(s_copy)
        ]

        return m


class LongDivider(Component):  # noqa: DOC602,DOC603
    r"""Long-division soft-core, used as a reference.

    This divider core is a gateware implementation of classic long division.

    * A divide starts on the current cycle when both ``inp.valid`` and
      ``inp.ready`` are asserted.
    * A divide result is available when ``outp.valid`` is asserted. The
      result is read/overwritten once a downstream core asserts ``outp.ready``.

    .. warning::

        This core is not intended to be used in user designs due to poor
        resource usage. It is mainly kept as a known-to-work reference design
        for possible equivalence checking later. Use :class:`MulticycleDiv`
        instead.

    * Latency: Divide Results for a will be available ``width`` clock cycles
      after assertion of both ``inp.valid`` and ``inp.ready``.

    * Throughput: One divide maximum is finished every ``width``
      clock cycles.

    Parameters
    ----------
    width : int
        Width in bits of both inputs ``n`` and ``d``. For signed
        multiplies, this includes the sign bit. Outputs ``q`` and ``r`` width
        will be the same width.

    Attributes
    ----------
    width : int
        Bit width of the inputs ``n`` and ``d``, and outputs ``q`` and ``r``.
    inp : In(divider_input_signature(width))
        Input interface to the divider.
    outp : Out(divider_output_signature(width))
        Output interface of the divider.

    Notes
    -----
    * This divider is a naive long-division implementation, similar to how
      pen-and-pad division is done in base-2/10.

    * Internally, the divider is aware of the sign of its inputs as well as
      the sign of the divide operation.

      The divider will dispatch to one of the 4 possible combinations of signs
      during the division loop. The four combinations are:

        * Add shifted copies of **positive**/**negative** powers of two
          together to form the quotient.
        * **Add**/**subtract** shifted copies of the divisor from the dividend
          **towards zero** to form the remainder.

    * This divider is compliant with RISC-V semantics for divide-by-zero and
      overflow when ``width=32`` or ``width=64``\ [rv]_. Specifically:

        * Signed/unsigned divide-by-zero returns "all bits set" for the
          quotient and the dividend as the remainder.
        * Signed overflow (:math:`-2^{\{31,63\}}/-1`) returns
          :math:`-2^{\{31,63\}}` for the quotient and :math:`0` as the
          remainder.

    .. [rv] https://github.com/riscv/riscv-isa-manual/releases/tag/Ratified-IMAFDQC
    """

    def __init__(self, width=8):
        self.width = width
        super().__init__({
            "inp": In(divider_input_signature(self.width)),
            "outp": Out(divider_output_signature(self.width))
        })

    def elaborate(self, platform):  # noqa: D102
        m = Module()

        m.submodules.mag = mag = _MagnitudeComparator(2*self.width)

        quotient = Signal(2*self.width)
        remainder = Signal(2*self.width)
        iters_left = Signal(range(self.width))
        in_progress = Signal()
        s_copy = Signal(Sign)
        a_sign = Signal()
        n_sign = Signal()

        # Reduce latency by 1 cycle by preempting output being read.
        m.d.comb += self.inp.ready.eq((self.outp.ready & self.outp.valid) |
                                      ~in_progress)
        m.d.comb += in_progress.eq(iters_left != 0)
        m.d.comb += [
            self.outp.payload.q.eq(quotient),
            self.outp.payload.r.eq(remainder),
            self.outp.payload.sign.eq(s_copy),
        ]

        with m.If(self.outp.ready & self.outp.valid):
            m.d.sync += self.outp.valid.eq(0)

        with m.If(self.inp.ready & self.inp.valid):
            m.d.sync += [
                # We handle first cycle using shift_amt mux.
                iters_left.eq(self.width - 1),
                a_sign.eq(self.inp.payload.n[-1]),
                n_sign.eq(self.inp.payload.d[-1]),
                s_copy.eq(self.inp.payload.sign)
            ]

            # When dividing by 0, for RISCV compliance, we need the division to
            # return -1. Without redirecting quotient calculation to the
            # "both dividend/divisor positive" case, dividing a negative
            # number by 0 returns 1. Note that the sign-bit processing doesn't
            # need to be special-cased, I do it anyway in case I see some
            # patterns I can refactor out.
            with m.If((self.inp.payload.sign == Sign.SIGNED) &
                      (self.inp.payload.d == 0) &
                      self.inp.payload.n[-1]):
                m.d.sync += a_sign.eq(0)
                m.d.sync += n_sign.eq(0)

            # We are committed to do a calc at this point, so might as well
            # start it now.
            #
            # Division quick refresh:
            #
            # Dividend / Divisor = Quotient
            # Dividend % Divisor = Remainder (takes the sign of the Dividend)
            #
            #         Quotient
            # Divisor)Dividend
            #         Remainder
            #
            # We're doing binary long division in hardware.
            # If starting with a positive dividend, we want to subtract
            # positive numbers from the dividend until we get as close to 0
            # without going below.
            # If '' negative dividend '' subtract negative numbers ''  as close
            # to 0 without going above. This becomes the remainder.
            #
            # We create the numbers to subtract from the dividend by
            # multiplying the divisor by either positive or negative powers
            # of two (shifting), depending on the sign of both the dividend
            # and divisor. When we add powers all of these two together, we
            # get the quotient.
            #
            # Either way, we can only subtract a shifted divisor from the
            # dividend if:
            #
            # * divisor*2**n <= dividend if dividend is positive.
            # * -divisor*2**n >= dividend if dividend is negative.
            #
            # We can use a magnitude comparator for this.

            shift_amt = 2**(self.width - 1)
            with m.If(self.inp.payload.sign == Sign.SIGNED):
                m.d.comb += [
                    mag.divisor.eq((self.inp.payload.d.as_signed() * shift_amt).as_signed()),  # noqa: E501
                    mag.dividend.eq(self.inp.payload.n.as_signed())
                ]
            with m.Else():
                m.d.comb += [
                    mag.divisor.eq(self.inp.payload.d.as_unsigned() *
                                   shift_amt),
                    mag.dividend.eq(self.inp.payload.n.as_unsigned())
                ]

            with m.If(mag.o):
                # If dividend/divisor are positive, subtract a positive
                # shifted divisor from dividend.
                with m.If((self.inp.payload.sign == Sign.UNSIGNED) |
                          (~self.inp.payload.n[-1] & ~self.inp.payload.d[-1]) |
                          (self.inp.payload.n[-1] & self.inp.payload.d == 0)):
                    m.d.sync += quotient.eq(C(1) * shift_amt)
                    with m.If(self.inp.payload.sign == Sign.SIGNED):
                        # If high bit is set, and, a signed div,
                        # we want to sign-extend.
                        m.d.sync += remainder.eq(self.inp.payload.n.as_signed() -  # noqa: E501
                                                 (self.inp.payload.d.as_signed() * C(1) * shift_amt).as_signed())  # noqa: E501
                    with m.Else():
                        # Otherwise, zero-extend.
                        m.d.sync += remainder.eq(self.inp.payload.n.as_unsigned() -  # noqa: E501
                                                 (self.inp.payload.d.as_unsigned() * C(1) * shift_amt))  # noqa: E501
                # If dividend is negative, but divisor is positive, create a
                # negative shifted divisor and subtract from the dividend.
                with m.If((self.inp.payload.sign == Sign.SIGNED) &
                          self.inp.payload.n[-1] & ~self.inp.payload.d[-1] &
                          ~(self.inp.payload.d == 0 & self.inp.payload.n[-1])):
                    m.d.sync += quotient.eq(C(-1) * shift_amt)
                    m.d.sync += remainder.eq(self.inp.payload.n.as_signed() -
                                             (self.inp.payload.d.as_signed() * C(-1) * shift_amt).as_signed())  # noqa: E501
                # If dividend is positive, but divisor is negative, create a
                # positive shifted divisor and subtract from the dividend.
                with m.If((self.inp.payload.sign == Sign.SIGNED) &
                          ~self.inp.payload.n[-1] & self.inp.payload.d[-1]):
                    m.d.sync += quotient.eq(C(-1) * shift_amt)
                    m.d.sync += remainder.eq(self.inp.payload.n.as_signed() -
                                             (self.inp.payload.d.as_signed() * C(-1) * shift_amt).as_signed())  # noqa: E501
                # If dividend/divisor is negative, subtract a negative
                # shifted divisor and subtract from the dividend.
                with m.If((self.inp.payload.sign == Sign.SIGNED) &
                          self.inp.payload.n[-1] & self.inp.payload.d[-1]):
                    m.d.sync += quotient.eq(C(1) * shift_amt)
                    m.d.sync += remainder.eq(self.inp.payload.n.as_signed() -
                                             (self.inp.payload.d.as_signed() * C(1) * shift_amt).as_signed())  # noqa: E501
            with m.Else():
                m.d.sync += quotient.eq(0)
                with m.If(self.inp.payload.sign == Sign.SIGNED):
                    # If high bit is set, and, a signed div,
                    # we want to sign-extend.
                    m.d.sync += remainder.eq(self.inp.payload.n.as_signed())
                with m.Else():
                    # Otherwise, zero-extend.
                    m.d.sync += remainder.eq(self.inp.payload.n.as_unsigned())

        # Main division loop.
        with m.If(in_progress):
            m.d.sync += iters_left.eq(iters_left - 1)

            shift_amt = (1 << (iters_left - 1).as_unsigned())
            with m.If(self.inp.payload.sign == Sign.SIGNED):
                m.d.comb += [
                    mag.divisor.eq(self.inp.payload.d.as_signed() * shift_amt),
                    mag.dividend.eq(remainder.as_signed())
                ]
            with m.Else():
                m.d.comb += [
                    mag.divisor.eq(self.inp.payload.d.as_unsigned() * shift_amt),  # noqa: E501
                    mag.dividend.eq(remainder)
                ]

            with m.If(mag.o):
                # If dividend/divisor are positive, subtract a positive
                # shifted divisor from dividend.
                with m.If((s_copy == Sign.UNSIGNED) | (~a_sign & ~n_sign)):
                    m.d.sync += quotient.eq(quotient + C(1) * shift_amt)
                    with m.If(s_copy == Sign.SIGNED):
                        # If high bit is set, and, a signed div,
                        # we want to sign-extend.
                        m.d.sync += remainder.eq(remainder -
                                                 (self.inp.payload.d.as_signed() * C(1) * shift_amt).as_signed())  # noqa: E501
                    with m.Else():
                        # Otherwise, zero-extend.
                        m.d.sync += remainder.eq(remainder -
                                                 (self.inp.payload.d.as_unsigned() * C(1) * shift_amt))  # noqa: E501
                # If dividend is negative, but divisor is positive, create a
                # negative shifted divisor and subtract from the dividend.
                with m.If((s_copy == Sign.SIGNED) & a_sign & ~n_sign): 
                    m.d.sync += quotient.eq(quotient + C(-1) * shift_amt)
                    m.d.sync += remainder.eq(remainder -
                                             (self.inp.payload.d.as_signed() * C(-1) * shift_amt).as_signed())  # noqa: E501
                # If dividend is positive, but divisor is negative, create a
                # positive shifted divisor and subtract from the dividend.
                with m.If((s_copy == Sign.SIGNED) & ~a_sign & n_sign):
                    m.d.sync += quotient.eq(quotient + C(-1) * shift_amt)
                    m.d.sync += remainder.eq(remainder -
                                             (self.inp.payload.d.as_signed() * C(-1) * shift_amt).as_signed())  # noqa: E501
                # If dividend/divisor are negative, subtract a negative
                # shifted divisor from dividend.
                with m.If((s_copy == Sign.SIGNED) & a_sign & n_sign):
                    m.d.sync += quotient.eq(quotient + C(1) * shift_amt)
                    m.d.sync += remainder.eq(remainder -
                                             (self.inp.payload.d.as_signed() * C(1) * shift_amt).as_signed())  # noqa: E501

            with m.If(iters_left - 1 == 0):
                m.d.sync += self.outp.valid.eq(1)

        return m


class _MagnitudeComparator(Elaboratable):
    """Compare the magnitude of two signed numbers."""

    def __init__(self, width=8):  # noqa: DOC1
        self.width = width
        self.dividend = Signal(signed(self.width))
        self.divisor = Signal(signed(self.width))
        self.o = Signal()

    def elaborate(self, platform):
        m = Module()

        m.d.comb += self.o.eq(abs(self.divisor) <= abs(self.dividend))

        return m



# amaranth: UnusedElaboratable=no

import pytest
from math import fmod
from smolarith.div import Sign, LongDivider, MulticycleDiv, Impl
import random


@pytest.fixture
def basic_testbench(mod, values, delay):
    m = mod

    async def testbench(ctx):
        (n, d, q, r, sign) = values
        await ctx.tick()

        ctx.set(m.inp.payload.n, n)
        ctx.set(m.inp.payload.d, d)
        ctx.set(m.inp.payload.sign, sign)
        ctx.set(m.inp.valid, 1)
        await ctx.tick()

        ctx.set(m.inp.valid, 0)  # Only schedule one xfer.
        ctx.set(m.outp.ready, 1)  # Immediately ready for retrieval.
        await ctx.tick().repeat(delay + 1)

        assert ctx.get(m.outp.valid) == 1
        assert ctx.get(m.outp.payload.sign) == sign.value

        if sign == Sign.UNSIGNED:
            assert ctx.get(m.outp.payload.q) == q
            assert ctx.get(m.outp.payload.r) == r
        else:
            assert ctx.get(m.outp.payload.q.as_signed()) == q
            assert ctx.get(m.outp.payload.r.as_signed()) == r

        await ctx.tick()
        assert ctx.get(m.outp.valid) == 0

    return testbench


@pytest.fixture(params=[Sign.UNSIGNED, Sign.SIGNED],
                ids=["i", "u"])
def all_values_tb(request, mod, delay):
    m = mod
    sign = request.param

    if sign == Sign.SIGNED:
        async def tb(ctx):
            await ctx.tick()

            for n in range(-2**(m.width-1), 2**(m.width-1)):
                ctx.set(m.inp.payload.n, n)
                for d in range(-2**(m.width-1), 2**(m.width-1)):
                    ctx.set(m.inp.payload.d, d)
                    ctx.set(m.inp.payload.sign, Sign.SIGNED)
                    ctx.set(m.inp.valid, 1)
                    await ctx.tick()

                    ctx.set(m.inp.valid, 0)  # Only schedule one xfer.
                    ctx.set(m.outp.ready, 1)  # Immediately ready for retrieval.  # noqa: E501
                    await ctx.tick().repeat(delay + 1)

                    assert ctx.get(m.outp.valid) == 1
                    assert ctx.get(m.outp.payload.sign) == Sign.SIGNED.value
                    if n == -2**(m.width-1) and d == -1:
                        assert ctx.get(m.outp.payload.q.as_signed()) == \
                            -2**(m.width-1)
                        assert ctx.get(m.outp.payload.r.as_signed()) == 0
                    elif d == 0:
                        assert ctx.get(m.outp.payload.q.as_signed()) == -1
                        assert ctx.get(m.outp.payload.r.as_signed()) == n
                    else:
                        assert ctx.get(m.outp.payload.q.as_signed()) == \
                            int(n / d)
                        assert ctx.get(m.outp.payload.r.as_signed()) == \
                            fmod(n, d)

                    await ctx.tick()
    else:  # Sign.UNSIGNED
        async def tb(ctx):
            await ctx.tick()

            for n in range(0, 2**m.width):
                ctx.set(m.inp.payload.n, n)
                for d in range(0, 2**m.width):
                    ctx.set(m.inp.payload.d, d)
                    ctx.set(m.inp.payload.sign, Sign.UNSIGNED)
                    ctx.set(m.inp.valid, 1)
                    await ctx.tick()

                    ctx.set(m.inp.valid, 0)  # Only schedule one xfer.
                    ctx.set(m.outp.ready, 1)  # Immediately ready for retrieval.  # noqa: E501
                    await ctx.tick().repeat(delay + 1)

                    assert ctx.get(m.outp.valid) == 1
                    assert ctx.get(m.outp.payload.sign) == Sign.UNSIGNED.value
                    if d == 0:
                        assert ctx.get(m.outp.payload.q.as_unsigned()) == \
                            2**m.width - 1
                        assert ctx.get(m.outp.payload.r.as_unsigned()) == n
                    else:
                        assert ctx.get(m.outp.payload.q.as_unsigned()) == \
                            int(n / d)
                        assert ctx.get(m.outp.payload.r.as_unsigned()) == \
                            fmod(n, d)

                    await ctx.tick()
    
    return tb


@pytest.fixture
def random_vals(mod):
    w = mod.width -1

    def shift_to_unsigned(v):
        if v == -2*w:
            return 0  # Most negative value maps to 0.
        elif v < 0:
            return v * -1
        else:  # v >= 0
            return v | (1 << w)

    def vals():
        for i in range(256):
            n = random.randint(-2**w, (2**w)-1)
            d = random.randint(-2**w, (2**w)-1)
            s = random.choice([Sign.UNSIGNED, Sign.SIGNED])

            if s == Sign.UNSIGNED:
                n = shift_to_unsigned(n)
                d = shift_to_unsigned(d)

                assert n >= 0
                assert d >= 0

            yield (n, d, s)

    return vals()


@pytest.fixture
def random_values_tb(mod, random_vals, delay):
    m = mod

    async def tb(ctx):
        await ctx.tick()

        for (n, d, s) in random_vals:
            ctx.set(m.inp.payload.n, n)
            ctx.set(m.inp.payload.d, d)
            ctx.set(m.inp.payload.sign, s)
            ctx.set(m.inp.valid, 1)
            await ctx.tick()

            ctx.set(m.inp.valid, 0)  # Only schedule one xfer.
            ctx.set(m.outp.ready, 1)  # Immediately ready for retrieval.  # noqa: E501
            await ctx.tick().repeat(delay + 1)

            assert ctx.get(m.outp.valid) == 1
            assert ctx.get(m.outp.payload.sign) == s.value

            if s == Sign.SIGNED:
                if n == -2**(m.width-1) and d == -1:
                    assert ctx.get(m.outp.payload.q.as_signed()) == \
                        -2**(m.width-1)
                    assert ctx.get(m.outp.payload.r.as_signed()) == 0
                elif d == 0:
                    assert ctx.get(m.outp.payload.q.as_signed()) == -1
                    assert ctx.get(m.outp.payload.r.as_signed()) == n
                else:
                    assert ctx.get(m.outp.payload.q.as_signed()) == \
                        int(n / d)
                    assert ctx.get(m.outp.payload.r.as_signed()) == \
                        fmod(n, d)
            else:
                if d == 0:
                    assert ctx.get(m.outp.payload.q.as_unsigned()) == \
                        2**m.width - 1
                    assert ctx.get(m.outp.payload.r.as_unsigned()) == n
                else:
                    assert ctx.get(m.outp.payload.q.as_unsigned()) == \
                        int(n / d)
                    assert ctx.get(m.outp.payload.r.as_unsigned()) == \
                        fmod(n, d)

            await ctx.tick()

    return tb


DIV_IDS=["nr", "res", "long"]


@pytest.mark.parametrize("mod,delay", [
    (MulticycleDiv(12, impl=Impl.NON_RESTORING), 15 - 1),
    (MulticycleDiv(12, impl=Impl.RESTORING), 14 - 1),
    (LongDivider(12), 11 - 1)],
    ids=DIV_IDS)
@pytest.mark.parametrize("clks", [1.0 / 12e6])
@pytest.mark.parametrize("values", [(1362, 14, 97, 4, Sign.SIGNED),
                                    (-1362, 14, -97, -4, Sign.SIGNED),
                                    (1362, -14, -97, 4, Sign.SIGNED),
                                    (-1362, -14, 97, -4, Sign.SIGNED),
                                    (1362, 14, 97, 4, Sign.UNSIGNED)],
                         ids=["i1", "i2", "i3", "i4", "u"])
def test_reference_div(sim, basic_testbench):
    sim.run(testbenches=[basic_testbench])


@pytest.mark.parametrize("mod,delay", [
    (MulticycleDiv(32, impl=Impl.NON_RESTORING), 35 - 1),
    (MulticycleDiv(32, impl=Impl.RESTORING), 34 - 1),
    (LongDivider(32), 31 - 1)],
    ids=DIV_IDS)
@pytest.mark.parametrize("clks", [1.0 / 12e6])
@pytest.mark.parametrize("values", [(-(2**31), -1, -(2**31), 0, Sign.SIGNED),
                                    (1, 0, -1, 1, Sign.SIGNED),
                                    (-1, 0, -1, -1, Sign.SIGNED),
                                    (0xff, 0, 2**32 - 1, 0xff, Sign.UNSIGNED)],
                         ids=["ov", "zero1", "zero2", "zero3"])
def test_riscv_compliance(sim, basic_testbench):
    sim.run(testbenches=[basic_testbench])


@pytest.mark.parametrize("mod,delay", [
    (MulticycleDiv(12, impl=Impl.NON_RESTORING), 15 - 1),
    (MulticycleDiv(12, impl=Impl.RESTORING), 14 - 1),
    (LongDivider(12), 11 - 1)],
    ids=DIV_IDS)
@pytest.mark.parametrize("clks", [1.0 / 12e6])
@pytest.mark.parametrize("values", 
                         [  
                             # 100000000001
                             (2049, 2, 1024, 1, Sign.UNSIGNED),
                             # 100000000001
                             (-2047, 2, -1023, -1, Sign.SIGNED),
                             # 100000000000
                             (2048, 2, 1024, 0, Sign.UNSIGNED),
                             # 100000000000
                             (-2048, 2, -1024, 0, Sign.SIGNED),
                             # 100000000000
                             (2048, 1, 2048, 0, Sign.UNSIGNED),
                             # 100000000000
                             (-2048, 1, -2048, 0, Sign.SIGNED),
                             # 011111111111
                             (2047, 2, 1023, 1, Sign.UNSIGNED),
                             # 011111111111
                             (2047, 2, 1023, 1, Sign.SIGNED)
                         ],
                         ids=["u1", "i1", "u2", "i2", "u3", "i3", "u4", "i4"])
def test_signed_unsigned_mismatch(sim, basic_testbench):
    sim.run(testbenches=[basic_testbench])


# FIXME: Use of fmod to calculate remainder causes precision issues for
# testing 64-bit dividers.
@pytest.mark.parametrize("mod,delay", [
    (MulticycleDiv(32, impl=Impl.NON_RESTORING), 35 - 1),
    (MulticycleDiv(32, impl=Impl.RESTORING), 34 - 1),
    (LongDivider(32), 31 - 1),
    pytest.param(MulticycleDiv(64, impl=Impl.NON_RESTORING), 67 - 1,
                 marks=pytest.mark.xfail(reason="fmod precision issues")),
    pytest.param(MulticycleDiv(64, impl=Impl.RESTORING), 66 - 1,
                 marks=pytest.mark.xfail(reason="fmod precision issues")),
    pytest.param(LongDivider(64), 63 - 1,
                 marks=pytest.mark.xfail(reason="fmod precision issues"))],
    ids=["n32", "r32", "l32", "n64", "r64", "l64"])
@pytest.mark.parametrize("clks", [1.0 / 12e6])
def test_random(sim, random_values_tb):
    random.seed(0)
    sim.run(testbenches=[random_values_tb])


@pytest.mark.parametrize("mod,delay", [
    (MulticycleDiv(8, impl=Impl.NON_RESTORING), 11 - 1),
    (MulticycleDiv(8, impl=Impl.RESTORING), 10 - 1),
    (LongDivider(8), 7 - 1)],
    ids=DIV_IDS)
@pytest.mark.parametrize("clks", [1.0 / 12e6]) 
def test_all_values(sim, all_values_tb):
    sim.run(testbenches=[all_values_tb])
